"""Plugin system core module
"""
from skippy.utils.logger import log

import skippy.config

from typing import Iterator, Optional, Callable, Union, Dict, Any
import importlib
import sys
import os

sys.path.append(skippy.config.PLUGINS_FOLDER)


def inject(
    module: str, attr: str, value: Optional[Any] = None
) -> Union[Any, Callable[[Any], Any]]:
    """Inject value into selected module attribute.

    Note: This function is not safe. It can be used only in plugins for injecting variable into skippy source code.

    Args:
        module (str): Module name
        attr (str): Attribute name
        value (Optional[Any], optional): Any value or None (if None func return decorator wrapper)

    Returns:
        Union[Any, Callable[Any]]: If value is not None return value, else return decorator wrapper
    """

    def inject(value: Any) -> Any:
        """Inject value into selected module attribute.

        Args:
            value (Any): Any value

        Returns:
            Any: Input value
        """
        loc = {}
        exec(f"import {module}; module = {module}", globals(), loc)
        setattr(loc["module"], attr, value)

        return value

    return inject(value) if value else inject


class BasePlugin:

    """Base plugin class"""

    __alias__ = "Plugin"

    __description__ = ""
    __author__ = ""
    __version__ = "1.0.0"

    def start(self):
        """Base start method"""
        pass

    def stop(self):
        """Base stop method"""
        pass


class PluginLoader:

    """Plugin loader class"""

    def __init__(self):
        """Init Plugin loader"""
        self._plugins = []

    @staticmethod
    def plugins() -> list:
        """Get list of all plugins

        Returns:
            list: List of plugins
        """
        return [
            os.path.splitext(file)[0]
            for file in os.listdir(skippy.config.PLUGINS_FOLDER)
            if os.path.isfile(os.path.join(skippy.config.PLUGINS_FOLDER, file))
            if file.endswith("_plugin.py")
        ]

    @classmethod
    def pluginsData(cls) -> Iterator[Dict[str, str]]:
        """Get data of all plugins

        Yields:
            Dict[str, str]: Dict of plugin data
        """
        for file in cls.plugins():
            plugin = cls.importPlugin(file)
            yield {
                "__alias__": plugin.__alias__,
                "__description__": plugin.__description__,
                "__author__": plugin.__author__,
                "__version__": plugin.__version__,
            }

    def loadPlugins(self):
        """Load all plugins"""
        for file in self.plugins():
            plugin = self.importPlugin(file)
            self._plugins.append(plugin)

    def startPlugins(self):
        """Start all plugins"""
        for plugin in self._plugins:
            plugin.start()
            log.debug(f"{plugin.__alias__} was started")

    def stopPlugins(self):
        """Stop all plugins"""
        for plugin in self._plugins:
            plugin.stop()
            log.debug(f"{plugin.__alias__} was stoped")

    @staticmethod
    def importPlugin(module: str) -> BasePlugin:
        """Import plugin by module name

        Args:
            module (str): Module name

        Returns:
            BasePlugin: Plugin instance
        """
        return importlib.import_module(module).load()



"""Summary
"""
from skippy.api import PageData

from skippy.utils.logger import log

from requests.exceptions import RequestException
from abc import ABCMeta, abstractmethod
from typing import Union, Tuple, List
import unicodedata
import pyftml
import pyscp
import html
import re


def render(pdata: PageData) -> str:
    """Summary

    Args:
        pdata (PageData): Description

    Returns:
        str: Description
    """
    preprocess = PreProcessorsHandler(pdata).process()
    htmlprocess = HTMLProcessorsHandler(preprocess, pdata).process()
    postprocess = PostProcessorsHandler(htmlprocess, pdata).process()

    return postprocess


#################################################
# Base Classes
#################################################
class AbstractProcessor(metaclass=ABCMeta):

    """Summary"""

    pattern: str

    def __init__(self, source: str, pdata: PageData):
        """Summary

        Args:
            source (str): Description
            pdata (PageData): Description
        """
        self.source: str = source
        self.pdata: PageData = pdata

    @property
    def matches(self) -> List[Union[Tuple[str, ...], str]]:
        """Summary

        Returns:
            List[Union[Tuple[str, ...], str]]: Description
        """
        return re.findall(
            self.pattern,
            self.source,
        )

    @abstractmethod
    def process(self):
        """Summary"""
        pass


class ProcessorsHandler(metaclass=ABCMeta):

    """Summary

    Attributes:
        source (str): Description
    """

    def __init__(self, source: str, pdata: PageData):
        """Summary

        Args:
            source (str): Description
            pdata (PageData): Description
        """
        self.source: str = source
        self.pdata: PageData = pdata
        self.processors: List[AbstractProcessor] = []

    def register(self, processor: AbstractProcessor):
        """Summary

        Args:
            processor (AbstractProcessor): Description
        """
        self.processors.append(processor)

    def process(self) -> str:
        """Summary

        Returns:
            str: Description
        """
        for processor in self.processors:
            try:
                self.source = processor(self.source, self.pdata).process()
            except Exception as e:
                log.error(e, exc_info=True)
        return self.source


#################################################
# ProcessorsHandlers
#################################################
class PreProcessorsHandler(ProcessorsHandler):

    """Summary"""

    def __init__(self, pdata: PageData):
        """Summary

        Args:
            pdata (PageData): Description
        """
        super(PreProcessorsHandler, self).__init__(pdata["source"], pdata)
        self.register(IncludesProcessor)
        self.register(IftagsProcessor)


class HTMLProcessorsHandler(ProcessorsHandler):

    """Summary

    Attributes:
        html (str): Description
    """

    def __init__(self, source: str, pdata: PageData):
        """Summary

        Args:
            source (str): Description
            pdata (PageData): Description
        """
        super(HTMLProcessorsHandler, self).__init__(source, pdata)
        self.register(MarkdownProcessor)
        self.register(InsertDataProcessor)
        self.register(ModuleCSSProcessor)

    def process(self):
        """Summary

        Returns:
            TYPE: Description
        """
        self.html = self.processors[0](self.source, self.pdata).process()
        for processor in self.processors[1:]:
            try:
                self.html = processor(self.source, self.pdata, self.html).process()
            except Exception as e:
                log.error(e, exc_info=True)
        return self.html


class PostProcessorsHandler(ProcessorsHandler):

    """Summary"""

    def __init__(self, source: str, pdata: PageData):
        """Summary

        Args:
            source (str): Description
            pdata (PageData): Description
        """
        super(PostProcessorsHandler, self).__init__(source, pdata)
        self.register(LocalImagesProcessor)
        self.register(HTMLTagsProcessor)


#################################################
# PreProcessors
#################################################
class IncludesProcessor(AbstractProcessor):

    """Summary

    Attributes:
        source (str): Description
    """

    pattern: str = r"(\[\[include\s(?::.+?:|)(?:.+?:|)(?:.+)(?:\s((?:.|\n)+?)|)]])"

    def process(self, iteration: int = 0) -> str:
        """Summary

        Returns:
            str: Description

        Args:
            iteration (int, optional): Description
        """
        for include in self.matches:
            path = re.findall(
                r"(?::(.+?):|)((?:.+?:|).+)",
                include[0].split()[1].replace("]]", ""),
            )[0]
            site = path[0]
            page = path[1]
            try:
                wiki = pyscp.wikidot.Wiki(site)
                p = wiki(page).source
                args = [
                    i[1:] if i.startswith("\n") else i for i in include[1].split("|")
                ]
                if not args[0]:
                    args = []
                for arg in args:
                    arg = re.findall(
                        r"(?:([\w-]+)(?:(?:\s|)=(?:\s|))((?:.|\n+?)+))",
                        arg,
                    )[0]
                    p = p.replace("{$" + arg[0] + "}", arg[1])
                self.source = self.source.replace(include[0], p)
            except Exception as e:
                log.error(e)
        if (
            re.findall(
                r"(\[\[include(?:(?:.|\n)+?)]])",
                self.source,
            )
            and iteration != 5
        ):
            self.process(iteration + 1)
        return self.source


class IftagsProcessor(AbstractProcessor):

    """Summary

    Attributes:
        source (str): Description
    """

    pattern: str = r"(\[\[iftags (.+?)]]((?:(?:.|\n|\s)+?))\[\[\/iftags]])"

    def process(self) -> str:
        """Summary

        Returns:
            str: Description
        """
        for iftag in self.matches:
            match = True
            elem = re.findall(r"(\+.+?(?=\s|$)|-.+?(?=\s|$))", iftag[1])
            for e in elem:
                if e.startswith("+"):
                    if e[1:] not in self.pdata["tags"]:
                        match = False
                else:
                    if e[1:] in self.pdata["tags"]:
                        match = False
            if match:
                self.source = self.source.replace(iftag[0], iftag[2])
        return self.source


#################################################
# HTMLProcessors
#################################################
class MarkdownProcessor(AbstractProcessor):

    """Summary"""

    wiki: pyscp.core.Wiki = pyscp.wikidot.Wiki("www.wikidot.com")

    def _wikidot(self) -> str:
        """Summary

        Returns:
            str: Description
        """
        return self.wiki._module("edit/PagePreviewModule", source=self.source)["body"]

    def _ftml(self) -> str:
        """Summary

        Returns:
            str: Description
        """
        return pyftml.render_html(self.source)["body"]

    def process(self) -> str:
        """Summary

        Returns:
            str: Description
        """
        try:
            return self._wikidot()
        except RequestException:
            return self._ftml()


class InsertDataProcessor(AbstractProcessor):

    """Summary"""

    html_base: str = """
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
        <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="ru" lang="ru" pdata-lt-installed="true">
            <head>
                <style type="text/css">
                    @import url(http://d3g0gp89917ko0.cloudfront.net/v--3e3a6f7dbcc9/common--theme/base/css/style.css);
                    @import url(http://scp-ru.wdfiles.com/local--code/component:theme2/1);
                </style>
                <script type="text/javascript" src="http://d3g0gp89917ko0.cloudfront.net/v--3e3a6f7dbcc9/common--javascript/init.combined.js"></script>
                <script type="text/javascript">
                    var URL_HOST = 'www.wikidot.com';
                    var URL_DOMAIN = 'wikidot.com';
                    var USE_SSL =  true ;
                    var URL_STATIC = 'http://d3g0gp89917ko0.cloudfront.net/v--3e3a6f7dbcc9';
                    // global request information
                    
                    var WIKIREQUEST = {};
                    WIKIREQUEST.info = {};
                    
                    WIKIREQUEST.info.domain = "www.wikidot.com";
                    WIKIREQUEST.info.siteId = 648902;
                    WIKIREQUEST.info.siteUnixName = "www";
                    WIKIREQUEST.info.categoryId = 4388020;
                    WIKIREQUEST.info.themeId = 192064;
                    WIKIREQUEST.info.requestPageName = "start:start";
                    OZONE.request.timestamp = 1619763822;
                    OZONE.request.date = new Date();
                    WIKIREQUEST.info.lang = 'en';
                            WIKIREQUEST.info.pageUnixName = "start:start";
                    WIKIREQUEST.info.pageId = 22129557;
                                    WIKIREQUEST.info.lang = "en";
                    OZONE.lang = "en";
                    var isUAMobile = !!/Android|webOS|iPhone|iPod|BlackBerry|IEMobile|Opera Mini/i.test(navigator.userAgent);
                </script>
                <script type="text/javascript">
            
                    require.config({
                        baseUrl: URL_STATIC + '/common--javascript',
                        paths: {
                            'jquery.ui': 'jquery-ui.min',
                            'jquery.form': 'jquery.form'
                        }
                    });
            
                </script>
                <script type="text/javascript" src="http://d3g0gp89917ko0.cloudfront.net/v--3e3a6f7dbcc9/common--javascript/WIKIDOT.combined.js"></script>
                <<MODULE-CSS-PREVIEW>>
                <style>
                    body {
                        padding: 2em !important;
                    }
                </style>
            </head>
            <body id="html-body">
                <div id="dummy-ondomready-block" style="display:none;"></div>
                <div id="page-title">
                    <<TITLE>>
                </div>
                <div id="page-content" style="min-height: 500px;">
                    <<CONTENT>>
                </div>
                <div class="page-tags">
                    <span>
                        <<TAGS>>
                    </span>
                </div>
            </body>
        </html>
        """

    def __init__(self, source: str, pdata: PageData, html: str):
        """Summary

        Args:
            source (str): Description
            pdata (PageData): Description
            html (str): Description
        """
        super(InsertDataProcessor, self).__init__(source, pdata)
        self.html: str = html

    def process(self) -> str:
        """Summary

        Returns:
            str: Description
        """
        self.source = self.html_base.replace(
            "<<TITLE>>", html.escape(self.pdata["title"])
        )
        self.source = self.source.replace("<<CONTENT>>", self.html)
        self.source = self.source.replace(
            "<<TAGS>>",
            " ".join([f"<a href='#'>{tag}</a>" for tag in self.pdata["tags"]]),
        )
        return self.source


class ModuleCSSProcessor(AbstractProcessor):

    """Summary

    Attributes:
        html (str): Description
    """

    pattern: str = r"(?:(?:\[\[module) (?:CSS|css)(?:(.+?)|)\]\]\n)((.|\n)+?)(?:\n\[\[\/module\]\])"

    def __init__(self, source: str, pdata: PageData, html: str):
        """Summary

        Args:
            source (str): Description
            pdata (PageData): Description
            html (str): Description
        """
        super(ModuleCSSProcessor, self).__init__(source, pdata)
        self.html: str = html

    def process(self) -> str:
        """Summary

        Returns:
            str: Description
        """
        styles = ""
        for style in self.matches:
            style = unicodedata.normalize("NFKD", style[1])
            styles += f"<style>\n{style}\n</style>\n"
        self.html = self.html.replace("<<MODULE-CSS-PREVIEW>>", styles)
        return self.html


#################################################
# PostProcessor
#################################################
class LocalImagesProcessor(AbstractProcessor):

    """Summary

    Attributes:
        source (str): Description
    """

    pattern: str = r'<img src="http://www.wdfiles.com/local--files//(.+?)"'

    def process(self) -> str:
        """Summary

        Returns:
            str: Description
        """
        files = self.pdata["files"]
        for img in self.matches:
            if img in files:
                self.source = self.source.replace(
                    f"http://www.wdfiles.com/local--files//{img}",
                    f"data:{img.split('.')[1]}/;base64,{files[img]}",
                )
        return self.source


class HTMLTagsProcessor(AbstractProcessor):

    """Summary

    Attributes:
        source (str): Description
    """

    pattern: str = r"(\[\[html(?:(?:.+?)| |)]]((?:.|\n|\s)+?)\[\[\/html]])"

    def process(self) -> str:
        """Summary

        Returns:
            str: Description
        """
        for htmltags in self.matches:
            self.source = self.source.replace(htmltags[0], html.unescape(htmltags[1]))
        return self.source



from PyQt5 import QtCore

from skippy.api import PageData

from skippy.core import scpclient

from skippy.gui import thread

from typing import List
import os


class FileWorker(thread.AbstractWorker):
    progress = QtCore.pyqtSignal(str, bytes)

    def __init__(self, files: List[str]):
        super(FileWorker, self).__init__()
        self.files = files

    def run(self):
        for file in self.files:
            if file:
                with open(file, "rb") as f:
                    self.progress.emit(os.path.split(file)[1], f.read())
        self.finished.emit()


class UploadWorker(thread.AbstractWorker):
    def __init__(self, pdata: PageData, comment: str = "Edit using Skippy"):
        super(UploadWorker, self).__init__()
        self.pdata = pdata
        self.comment = comment

    def run(self):
        scpclient.SCPClient().upload(self.pdata, self.comment)
        self.finished.emit()



from typing import NamedTuple, TypedDict, Callable, Optional, Tuple, Dict, List, Any


class Language(NamedTuple):
    code: str
    dictionary: Dict[str, str]


class Action(NamedTuple):
    label: str
    statusTip: str
    func: Callable[..., Any]
    img: str


class InlineSyntax(NamedTuple):
    opening: str
    closing: str
    multiple: bool


class BlockSyntax(NamedTuple):
    closeBlock: bool
    inline: bool
    dataVar: Dict[str, Dict[str, str]] = {"": {}}


class PageData(TypedDict):
    title: str
    source: str
    tags: List[str]
    files: Dict[str, str]
    link: Optional[Tuple[str, str]]


__all__ = ["Language", "Action", "InlineSyntax", "BlockSyntax", "PageData"]



"""Singleton facade for pyscp module.
"""
from skippy.api import ConnectionErrors, Singleton, PageData, critical, ignore

from urllib.parse import urlparse
from pyscp.wikidot import Wiki
from typing import Optional
import requests
import base64


class SCPClient(metaclass=Singleton):

    """Singleton facade for pyscp module."""

    def __init__(self, login: Optional[str] = None, password: Optional[str] = None):
        """Initialize SCPClient.

        Args:
            login (Optional[str], optional): User login
            password (Optional[str], optional): User password
        """
        self._session = None

        if login and password:
            self.auth(login, password)

    @critical
    @ignore(ConnectionErrors)
    def auth(self, login: str, password: str):
        """Retrieving Wikidot Session from Credentials.

        Args:
            login (str): User login
            password (str): User password
        """
        wikidot = Wiki("www.wikidot.com")
        wikidot.auth(login, password)

        self._session = wikidot.cookies

    @critical
    @ignore(ConnectionErrors)
    def getWiki(self, site: str) -> Wiki:
        """Get Wiki object instance by Wikidot site name with current session cookies.

        Args:
            site (str): Wikidot site name

        Returns:
            Wiki: Wiki object instance with current session cookies
        """
        wiki = Wiki(site)
        if self._session:
            wiki.cookies = self._session

        return wiki

    @critical
    @ignore(ConnectionErrors)
    def upload(self, page: PageData, comment: str = "Edit using Skippy"):
        """Upload page with tags and files to selected Wikidot site.

        Args:
            page (PageData): Page for given Wikidot site
            comment (str, optional): Comment for uploaded changes
        """
        wiki = self.getWiki(page["link"][0])
        p = wiki(page["link"][1])

        p.edit(title=page["title"], source=page["source"], comment=comment)

        p.set_tags(page["tags"])

        for file in page["files"]:
            file_source = base64.b64decode(page["files"][file])
            try:
                p.upload(file, file_source)
            except RuntimeError:
                try:
                    p.remove_file(file)
                    p.upload(file, file_source)
                except RuntimeError:
                    pass

        for file in p.files:
            name = file.name
            if name not in page["files"]:
                p.remove_file(name)

    @ignore(Exception, {})
    @ignore(ConnectionErrors)
    def download(self, site: str, page: str) -> PageData:
        """Download page with tags and files fron selected Wikidot site.

        Args:
            site (str): Wikidot site name
            page (str): Page for given Wikidot site

        Returns:
            PageData: Page data
        """
        wiki = self.getWiki(site)
        p = wiki(page)

        files = {}
        for file in p.files:
            with requests.get(file.url) as res:
                files[file.name] = base64.b64encode(res.content).decode("utf-8")

        return {
            "title": p.title,
            "source": p.source,
            "tags": list(p.tags),
            "files": files,
            "link": (urlparse(wiki.site).netloc, page),
        }



from PyQt5 import QtWidgets, QtCore

from skippy.api import ConnectionErrors, ignore

from skippy.gui import thread

from skippy.utils import translator, filehandlers

from typing import Optional, List
from urllib.parse import urlparse
import functools
import pyscp


class UserSitesWorker(thread.AbstractWorker):
    finished = QtCore.pyqtSignal(list)

    def run(self):
        self.finished.emit(self.sites)

    @functools.cached_property
    @ignore(ConnectionErrors, [])
    def sites(self) -> List[str]:
        return [
            urlparse(wiki.site).netloc
            for wiki in pyscp.wikidot.User(
                filehandlers.ProfileHandler().load()[0]
            ).member
        ]


class SiteBox(QtWidgets.QComboBox):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None, default: str = ""):
        super(SiteBox, self).__init__(parent)
        self.default = default

        self.lineEdit = QtWidgets.QLineEdit(self)
        self.lineEdit.setPlaceholderText(
            translator.Translator().translate("SITE_BOX_LINEEDIT")
        )

        self.setLineEdit(self.lineEdit)
        self.setEditable(True)

        self.setCurrentText(default)

        self._thread = thread.Thread(UserSitesWorker())
        self._thread.worker.finished.connect(self.addSites)
        self._thread.start()

    def addSites(self, sites: List[str]):
        self.addItems(sites)
        self.setCurrentText(self.default)
        self.setCompleter(QtWidgets.QCompleter(sites))
        if not self.currentText():
            self.lineEdit.clear()
